@@ -50,10 +50,10 @@ class Handler:
                 self.kwargs[param_name] = param.default
                 continue
             if isinstance(param.annotation, Task):
-                self.kwargs[param_name] = dependencies.Task()
+                self.kwargs[param_name] = dependencies.GetTask()
                 continue
             if isinstance(param.annotation, GeneratedProtocolMessageType):
-                self.kwargs[param_name] = dependencies.Payload(param.annotation)
+                self.kwargs[param_name] = dependencies.GetPayload(param.annotation)
                 continue
 
             raise NotImplementedError(f"{param_name} is not supported")
