@@ -3,18 +3,70 @@
 import asyncio
 import logging
 from datetime import datetime
-from functools import partial
-from typing import MutableMapping, List, Union, Coroutine, Type
+from functools import partial, lru_cache
+from itertools import chain
+from operator import itemgetter
+from typing import MutableMapping, List, Union, Coroutine, Type, Sequence, Dict, Callable, Awaitable, Optional, \
+    Iterator, NoReturn
 
 from patchwork.core import Task, AsyncSubscriber
 from patchwork.core.config.base import ComponentConfig
+from patchwork.core.typing import FuncPath
 from patchwork.core.utils import cached_property
 from patchwork.node.core import Module
-from patchwork.node.core.exceptions import TaskControlException, TaskRetry, TaskFatal, TaskDrop
+from patchwork.node.core.exceptions import TaskControlException, TaskRetry, TaskFatal, TaskDrop, TaskNoHandler
 from patchwork.node.core.router import TaskRouter
 from patchwork.node.executor.unit import InlineUnit
 from patchwork.node.executor.unit.base import ProcessingUnit
 
+# task finalizer, run when task processing completes, if third argument is None
+# means that task has been not processed as there was no processor to handle it
+FinalizerType = Callable[[Task, AsyncSubscriber, asyncio.Future], Awaitable]
+# middleware is to prepare task processing environment, optional finalizer is a method which
+# will be called to finalize task processing. Middleware may return new finalizer which should
+# invoke given finalizer internally for proper task finalization
+# if middleware does not return new finalizer, passed finalizer will be used
+# middlewares are run on the worker async loop, before passing task to processing unit!
+# keep this in mind as middleware execution may slow down whole executor
+MiddlewareType = Callable[[Task, AsyncSubscriber, FinalizerType], Awaitable[Optional[FinalizerType]]]
+
+ExceptionHandler = Callable[[Task, Exception], Awaitable]
+
+
+async def check_task_middleware(task: Task, receiver: AsyncSubscriber, finalizer: FinalizerType):
+    now = datetime.utcnow().timestamp()
+    if task.meta.HasField('expires') and now > task.meta.expires.ToSeconds():
+        # task has expired
+        raise TaskFatal(reason="Task has expired")
+    elif task.meta.HasField('not_before') and now < task.meta.not_before.ToSeconds():
+        # task is not ready yet
+        raise TaskRetry(countdown=(now - task.meta.not_before))
+
+
+async def default_task_exception_handler(task: Task, exception: Exception) -> NoReturn:
+    if isinstance(exception, TaskControlException):
+        raise exception
+
+    # request task retry
+    raise TaskRetry()
+
+
+class Idempotent:
+
+    idempotent = True
+
+    def __init__(self, fn):
+        self._fn = fn
+
+    def __call__(self, task: Task, exc: Exception):
+        try:
+            return self._fn(task, exc)
+        except TaskControlException:
+            raise RuntimeError(f"Idempotent exception handlers can't raise TaskControlExceptions")
+
+    def __repr__(self):
+        return f"idempotent<{repr(self._fn)}>"
+
 
 class BaseExecutor(Module):
     """
@@ -27,6 +79,9 @@ class BaseExecutor(Module):
         # executor termination timeout in seconds
         terminate_timeout: int = 10
         unit: ComponentConfig[Type[ProcessingUnit]] = ComponentConfig(InlineUnit)
+        middlewares: Sequence[FuncPath] = [
+            FuncPath(check_task_middleware)
+        ]
 
     logger_name = 'executor'
     _loop_task: asyncio.Future
@@ -39,10 +94,17 @@ class BaseExecutor(Module):
         self._terminate_request = asyncio.Event()
         self._current_tasks = set()
         self._routers = []
+        self._middlewares: List[MiddlewareType] = []
+        self._exc_handlers: Dict[type, List] = {}
+
+        for middleware in self.settings.middlewares:
+            self.add_middleware(middleware)
+
+        self.add_exception_handler(Exception, default_task_exception_handler)
 
     def __repr__(self):
         res = super().__repr__()
-        return f"<{res[1:-1]} [subscribers={'|'.join(self.settings.listen_on)} " \
+        return f"<{res[1:-1]} " \
                f"unit={self.unit.__class__.__name__}]>"
 
     def get_state(self) -> MutableMapping:
@@ -59,6 +121,31 @@ class BaseExecutor(Module):
         router.bind(self.worker)
         self._routers.append(router)
 
+    def add_middleware(self, middleware: MiddlewareType):
+        self._middlewares.append(middleware)
+
+    def add_exception_handler(self, exception: type, handler: ExceptionHandler, idempotent: bool = False):
+        """
+        Register a handler for given exception type (and subtypes). Exception handlers are called
+        in LIFO manner - last registered handler is firstly called.
+
+        If any handler raises one of TaskControlException it's handled by the executor and then no more exception
+        handlers are executed (except idempotent ones)
+        :param exception:
+        :param handler:
+        :param idempotent: if True, handler will be executed even if TaskControlException has been raised by one
+        of previously executed handlers.
+            :: warning
+                Idempotent handlers CAN'T raise TaskControlExceptions.
+        :return:
+        """
+        if exception not in self._exc_handlers:
+            self._exc_handlers[exception] = []
+
+        if idempotent:
+            handler = Idempotent(handler)
+        self._exc_handlers[exception].insert(0, handler)
+
     async def _start(self):
         """
         Setups executor
@@ -203,22 +290,21 @@ class BaseExecutor(Module):
             # log error and rollback task if something went wrong during internal scheduling
             self.logger.error(f"Unable to handle task '{task.uuid}', exception raised: {e.__class__.__name__}({e})")
             try:
-                await self.retry(task, receiver)
+                await self.retry(task)
+                # TODO: commit or not?
             except Exception as e:
                 self.logger.error(f"Unable to retry unhandled task '{task}', "
                                   f"exception raised: {e.__class__.__name__}({e})")
                 await self.backoff_task(task, reason="Internal retry failed")
+                # TODO: commit or not?
 
         return True
 
-    async def retry(self, task: Task, receiver: AsyncSubscriber = None, *, countdown: float = None,
+    async def retry(self, task: Task, *, countdown: float = None,
                     not_before: datetime = None, bump_attempts: bool = True) -> Union[Task, None]:
         """
-        Retries given task by cloning it and sending back to originating client saved in 'receiver' task argument.
-        If receiver is unknown task is send through default client using `send_task` method. When new task
-        is delivered to the queue old one is committed.
+        Retries given task by cloning it and sending through publisher.
         :param task:
-        :param receiver:
         :param countdown:
         :param not_before:
         :param bump_attempts:
@@ -230,7 +316,7 @@ class BaseExecutor(Module):
 
         if bump_attempts:
             new_task.meta.attempt += 1
-            if new_task.meta.attempt > new_task.meta.max_retries:
+            if task.meta.max_retries > 0 and new_task.meta.attempt > new_task.meta.max_retries:
                 await self.backoff_task(task, reason="Max number of retries exceeded")
                 return
 
@@ -242,9 +328,8 @@ class BaseExecutor(Module):
         if not_before is not None:
             new_task.meta.not_before.FromDatetime(not_before)
 
-        await self.worker.get_publisher('default').send_task(new_task)
-
-        await self._commit_task(task, receiver)
+        publisher = self.worker.get_publisher()
+        await publisher.send_task(new_task)
         return new_task
 
     async def backoff_task(self, task, reason: str = None):
@@ -283,120 +368,165 @@ class BaseExecutor(Module):
         :return: future of task processing function or None if processing was not scheduled because of
                  task control exception
         """
+        finalizer = self._handle_task_result
+        task_future = self.app_loop.create_future()
 
         try:
-            for router in self._routers:
-                processor = router.handle(task, receiver)
-                if processor is not None:
-                    task_future = await self._execute(task, receiver, processor)
-                    self._current_tasks.add(task_future)
-                    return task_future
+            for middleware in self._middlewares:
+                finalizer = (await middleware(task, receiver, finalizer)) or finalizer
 
-            self.logger.error(f"No processor to handle message type '{task.task_type}'")
-            raise TaskFatal(task, reason="No processor to handle this message type")
+            finalizer = partial(finalizer, task=task, receiver=receiver)
 
-        except TaskControlException as exc:
-            # allow _execute to raise task control exceptions
-            await self._handle_task_control(task, receiver, exc)
-            return None
+            exec_fut = await self._execute(task, receiver)
+        except Exception as exc:
+            exec_fut = self.app_loop.create_future()
+            exec_fut.set_exception(exc)
 
-    async def _execute(self, task: Task, receiver: AsyncSubscriber, processor: Coroutine):
+        exec_fut.add_done_callback(
+            partial(self._task_done_callback, finalizer=finalizer(processing_fut=exec_fut), task_future=task_future)
+        )
+
+        self._current_tasks.add(task_future)
+        task_future.add_done_callback(lambda f: self._current_tasks.remove(task_future))
+
+        return task_future
+
+    async def _execute(self, task: Task, receiver: AsyncSubscriber):
         """
         Executes a task on given processor
         :param task:
         :param receiver:
-        :param processor:
         :return:
         """
-        # verify if task is still valid
-        self.check_task(task)
+        for router in self._routers:
+            processor = router.handle(task, receiver)
+            if processor is not None:
+                # await for submit which may hold if processing unit is full
+                pu_fut = await self.unit.submit(processor)
+                return pu_fut
 
-        # await for submit which may hold if processing unit is full
-        fut: asyncio.Future = await self.unit.submit(processor)
-        fut.add_done_callback(partial(self._task_done_callback, receiver=receiver, task=task))
+        self.logger.error(f"No processor to handle message type '{task.task_type}'")
+        raise TaskNoHandler()
 
-        return fut
-
-    def check_task(self, task: Task):
-        """
-        Validates if given task is still valid and should be executed
-        :param task:
-        :return:
-        """
-        now = datetime.utcnow().timestamp()
-        if task.meta.HasField('expires') and now > task.meta.expires.ToSeconds():
-            # task has expired
-            self.logger.debug(f"Task '{task.uuid}' is not valid after: {task.meta.expires}, now: {now}")
-            raise TaskFatal(task, reason="Task has expired")
-        elif task.meta.HasField('not_before') and now < task.meta.not_before.ToSeconds():
-            # task is not ready yet
-            self.logger.debug(f"Task '{task.uuid}' is not valid before: {task.meta.not_before}, now: {now}")
-            raise TaskRetry(task, countdown=(now-task.meta.not_before))
-
-    def _task_done_callback(self, fut: asyncio.Future, receiver: AsyncSubscriber, task: Task):
+    def _task_done_callback(self,
+                            processing_fut: asyncio.Future,
+                            finalizer: Coroutine,
+                            task_future: asyncio.Future
+                            ):
         """
         Handles value returned by task processing future.
-        :param task:
-        :param receiver:
-        :param fut:
+        :param finalizer:
+        :param processing_fut:
+        :param task_future:
         :return:
         """
         def _callback(f: asyncio.Future):
-            self._current_tasks.remove(fut)
             if f.cancelled():
-                # task has been cancelled
+                # if finalizer has been cancelled to not pass result from task
+                # processing future to task future, set it to None as there is no ready result as finalizers
+                # were not completed
+                task_future.set_result(None)
                 return
 
             exc = f.exception()
             if exc is not None:
+                # if finalizer ends with unhandled exception it's internal fatal error
+                # kill the executor
+                task_future.set_result(None)
                 self.logger.fatal(f"Can't handle task result: {exc.__class__.__name__}({exc})")
-                self.harakiri()
+                self.harakiri(exc)
 
-        handle_fut = self.app_loop.create_task(self._handle_task_result(task, receiver, fut))
+            # otherwise pass processing future result to task result
+            if processing_fut.cancelled():
+                task_future.cancel()
+            elif processing_fut.exception():
+                task_future.set_exception(processing_fut.exception())
+            else:
+                task_future.set_result(processing_fut.result())
+
+        handle_fut = self.app_loop.create_task(finalizer)
         handle_fut.add_done_callback(_callback)
 
-    async def _handle_task_result(self, task: Task, receiver: AsyncSubscriber, task_fut: asyncio.Future):
+    @lru_cache(maxsize=None)
+    def _get_exception_handler(self, exception) -> Iterator[ExceptionHandler]:
+        handlers = []
+
+        for exc_type, callbacks in self._exc_handlers.items():
+            if not isinstance(exception, exc_type):
+                continue
+
+            done = False
+            for idx, item in enumerate(handlers):
+                if exc_type == item[0]:
+                    item[1].extend(callbacks)
+                    done = True
+                    break
+
+                if issubclass(exc_type, item[0]):
+                    handlers.insert(idx, (exc_type, list(callbacks)))
+                    done = True
+                    break
+
+            if not done:
+                handlers.append((exc_type, list(callbacks)))
+
+        return chain.from_iterable(map(itemgetter(1), handlers))
+
+    async def _handle_task_result(self, task: Task, receiver: AsyncSubscriber, processing_fut: asyncio.Future):
         """
         Handles task future result.
         :param task:
         :param receiver:
-        :param task_fut:
+        :param processing_fut:
         :return:
         """
-        exc = task_fut.exception()
-        if task_fut.cancelled():
+        exc = processing_fut.exception()
+        if processing_fut.cancelled():
             self.logger.debug(f"Task '{task}' execution cancelled internally")
-            await self.retry(task, receiver)
+            return
         elif exc is not None:
-            if isinstance(exc, TaskControlException):
-                await self._handle_task_control(task, receiver, exc)
-            else:
-                self.logger.warning(f"Task '{task}' execution failed with exception {exc.__class__.__name__}({exc})")
-                await self.retry(task, receiver)
-        else:
-            await self._commit_task(task, receiver)
+            self.logger.warning(f"Task '{task}' execution failed with exception {exc.__class__.__name__}({exc})")
+            handlers = self._get_exception_handler(exc)
+            finalized = False
 
-    async def _handle_task_control(self, task: Task, receiver: AsyncSubscriber, exc: TaskControlException):
+            for handler in handlers:
+                if finalized and not getattr(handler, 'idempotent', False):
+                    # if task is finalized (there was a TCE raised and handled), run only
+                    # handlers declared as idempotent
+                    continue
+
+                try:
+                    await handler(task, exc)
+                except TaskControlException as control_exc:
+                    await self._handle_task_control(task, control_exc)
+                    finalized = True
+                except Exception as exc:
+                    self.logger.exception(f"exception handler failed: {exc.__class__.__name__}({exc})")
+                    if self.worker.debug:
+                        raise exc
+
+        await self._commit_task(task, receiver)
+
+    async def _handle_task_control(self, task: Task, exc: Exception):
         """
         Takes appropriate action for the task depending on given task control exception.
         This method commits task.
         :param task:
-        :param receiver:
         :param exc:
         :return:
         """
         if isinstance(exc, TaskRetry):
             # schedule new task
-            await self.retry(task, receiver, not_before=exc.not_before)
+            await self.retry(task, not_before=exc.not_before)
         elif isinstance(exc, TaskFatal):
             await self.backoff_task(task, reason=(exc.reason or "Task processing fatal error"))
             # backoff just logs task, commit it to truly remove from the queue
-            await self._commit_task(task, receiver)
         elif isinstance(exc, TaskDrop):
-            await self._commit_task(task, receiver)
-        else:
+            pass
+        elif isinstance(exc, TaskControlException):
             self.logger.error(f"Unsupported task control exception '{exc.__class__.__name__}({exc})'")
-            raise ValueError(f"Unsupported task control exception: '{exc.__class__.__name__}({exc})'") from exc
+            if self.worker.debug:
+                raise ValueError("unknown task control exception")
 
     async def _commit_task(self, task: Task, receiver: AsyncSubscriber):
         """
@@ -406,7 +536,7 @@ class BaseExecutor(Module):
         """
         await receiver.commit(task)
 
-    def harakiri(self):
+    def harakiri(self, cause: BaseException = None):
         """
         Stops executor immediately due to critical internal error.
         Call this method if something went so wrong that there is no better way than killing this executor
@@ -414,5 +544,9 @@ class BaseExecutor(Module):
         are in progress or pending. Broker should consider this worker as dead and stop delivering new messages.
         :return:
         """
+        if self.worker.debug and cause:
+            # in debug mode raise SystemExit to blow up
+            raise SystemExit() from cause
+
         self.logger.error("Executor harakiri requested")
         self.app_loop.create_task(self.terminate())
